package assistedinject;

import com.google.inject.Inject;
import com.google.inject.TypeLiteral;

import java.lang.annotation.Annotation;
import java.lang.reflect.Constructor;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.List;

/**
 * Internal respresentation of a constructor annotated with
 * {@link AssistedInject}
 * 
 * @author <a href="jmourits@gmail.com">Jerome Mourits</a>
 * @author <a href="jesse@swank.ca">Jesse Wilson</a>
 */
class AssistedConstructor<T> {

  private final Constructor<T> constructor;
  private final ParameterListKey assistedParameters;
  private final List<Parameter> allParameters;

  @SuppressWarnings("unchecked")
  public AssistedConstructor(Constructor<T> constructor) {
    this.constructor = constructor;

    Type[] parameterTypes = constructor.getGenericParameterTypes();
    Annotation[][] annotations = constructor.getParameterAnnotations();

    List<TypeLiteral> assistedParameters = new ArrayList<TypeLiteral>();
    allParameters = new ArrayList<Parameter>();

    // categorize params as @Assisted or @Injected
    for (int i = 0; i < parameterTypes.length; i++) {
      Parameter parameter = new Parameter(parameterTypes[i], annotations[i]);
      allParameters.add(parameter);
      if (parameter.isProvidedByFactory()) {
        assistedParameters.add(parameter.getTypeLiteral());
      }
    }
    this.assistedParameters = new ParameterListKey(assistedParameters);
  }
  
  /**
   * Returns the {@link ParameterListKey} for this constructor.  The
   * {@link ParameterListKey} is created from the ordered list of {@link Assisted}
   * constructor parameters.
   */
  public ParameterListKey getAssistedParameters() {
    return assistedParameters;
  }
  
  /**
   * Returns an ordered list of all constructor parameters (both
   * {@link Assisted} and {@link Inject}ed).
   */
  public List<Parameter> getAllParameters() {
    return allParameters;
  }
  
  /**
   * Returns an instance of T, constructed using this constructor, with the
   * supplied arguments.
   */
  public T newInstance(Object[] args) {
    try {
      return constructor.newInstance(args);
    } catch (Exception e) {
      throw new RuntimeException("could not instantiate");
    }
  }
  
  @Override
  public String toString() {
    return constructor.toString();
  }
}
