package assistedinject;

import com.google.inject.Inject;
import com.google.inject.Injector;
import com.google.inject.Provider;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Provides a mechanism to combine user-specified paramters with
 * {@link Injector}-specified paramters when creating new objects.
 *
 * <p>To use a {@link FactoryProvider}:
 *
 * <p>Annotate your implementation class' constructor with the
 * {@link @AssistedInject} and the user-specified parameters with
 * {@link @Assisted}:
 * <pre><code>public class RealPayment implements Payment {
 *    &#64;AssistedInject
 *    public RealFoo(CreditService creditService, AuthService authService,
 *      &#64;Assisted Date startDate, &#64;Assisted Money amount) {
 *     ...
 *  }
 * }</code></pre>
 *
 * <p>Write an interface with a <i>create</i> method that accepts the user-specified
 * parameters in the same order as they appear in the implementation class' constructor:
 * <pre><code>public interface PaymentFactory {
 *    Payment create(Date startDate, Money amount);
 * }</code></pre>
 *
 * <p>You can name your create methods whatever you like, such as <i>create</i>,
 * or <i>createPayment</i> or <i>newPayment</i>. The implementation class must
 * be assignable to the return type of your create method. You can also provide
 * multiple factory methods, but there must be exactly one {@link @AssistedInject}
 * constructor on the implementation class for each.
 *
 * <p>In your Guice {@link com.google.inject.Module module}, bind your factory
 * interface to an instance of {@link FactoryProvider} that was created with
 * the same factory interface and implementation type:
 * <pre><code>  bind(FooFactory.class)
 *     .toProvider(FactoryProvider.newFactory(PaymentFactory.class, RealPayment.class));</code></pre>
 *
 * <p>Now you can {@code @Inject} your factory interface into your Guice-injected
 * classes. When you invoke the create method on that factory, the
 * {@link FactoryProvider} will instantiate the implementation class using
 * parameters from the injector and the factory method.
 *
 * <pre><code>public class PaymentAction {
 *    @Inject private PaymentFactory paymentFactory;
 *
 *    public void doPayment(Money amount) {
 *       Payment payment = paymentFactory.create(new Date(), amount);
 *       payment.apply();
 *    }
 * }
 * 
 * @author <a href="jmourits@gmail.com">Jerome Mourits</a>
 * @author <a href="jesse@swank.ca">Jesse Wilson</a>
 */
public class FactoryProvider<F, R> implements Provider<F> {

  private Injector injector;

  private final Class<F> factoryType;
  private final Class<R> implementationType;
  private final Map<Method, AssistedConstructor> factoryMethodToConstructor;

  public static <X,Y> FactoryProvider<X,Y> newFactory(
      Class<X> factoryType, Class<Y> implementationType){
    return new FactoryProvider<X, Y>(factoryType,implementationType);
  }
  
  private FactoryProvider(Class<F> factoryType, Class<R> implementationType) {
    this.factoryType = factoryType;
    this.implementationType = implementationType;
    this.factoryMethodToConstructor = createMethodMapping();
  }
  
  @Inject
  @SuppressWarnings({"unchecked", "unused"})
  private void setInjectorAndCheckUnboundParametersAreInjectable(
      Injector injector) {
    this.injector = injector;
    for (AssistedConstructor<?> c : factoryMethodToConstructor.values()) {
      for (Parameter p : c.getAllParameters()) {
        if (!p.isProvidedByFactory() && !paramCanBeInjected(p, injector)) {
          throw new RuntimeException(
              String.format("Parameter of type '%s' is not injectable or annotated with @Assisted for Constructor '%s'", p, c));
        }
      }
    }
  }
  
  @SuppressWarnings("unchecked")
  private boolean paramCanBeInjected(Parameter parameter, Injector injector) {
    return parameter.isBound(injector);
  }

  @SuppressWarnings({"unchecked"})
  private Map<Method, AssistedConstructor> createMethodMapping() {
    
    List<AssistedConstructor<?>> constructors = new ArrayList<AssistedConstructor<?>>();
    
    for (Constructor c : implementationType.getConstructors()) {
      if (c.getAnnotation(AssistedInject.class) != null) {
        constructors.add(new AssistedConstructor(c));
      }
    }
    
    if (constructors.size() != factoryType.getMethods().length) {
      throw new IllegalArgumentException("Constructor mismatch");
    }
    
    Map<ParameterListKey, AssistedConstructor> paramsToConstructor
        = new HashMap<ParameterListKey, AssistedConstructor>();
    
    for (AssistedConstructor c : constructors) {
      if (paramsToConstructor.containsKey(c.getAssistedParameters())) {
        throw new RuntimeException("Duplicate constructor, " + c);
      }
      paramsToConstructor.put(c.getAssistedParameters(), c);
    }
    
    Map<Method, AssistedConstructor> result = new HashMap<Method, AssistedConstructor>();
    for (Method method : factoryType.getMethods()) {
      if (implementationType.isAssignableFrom(method.getReturnType())) {
        throw new RuntimeException("Return method not assignable to blah");
      }
      ParameterListKey methodParams = new ParameterListKey(method.getGenericParameterTypes());
      
      if (!paramsToConstructor.containsKey(methodParams)) {
        throw new IllegalArgumentException(String.format("%s has no " +
            "@AssistInject constructor that takes the @Assisted parameters %s " +
            "in that order. @AssistInject constructors are %s",
            implementationType, methodParams, paramsToConstructor.values()));
      }
      AssistedConstructor matchingConstructor = paramsToConstructor.remove(methodParams);
      
      result.put(method, matchingConstructor);
    }
    return result;
  }

  @SuppressWarnings({"unchecked"})
  public F get() {
    InvocationHandler invocationHandler = new InvocationHandler() {

      public Object invoke(Object proxy, Method method, Object[] creationArgs) {
        AssistedConstructor<?> constructor = factoryMethodToConstructor.get(method);

        Object[] constructorArgs = gatherArgsForConstructor(
            constructor, creationArgs);
        Object objectToReturn = constructor.newInstance(constructorArgs);
        injector.injectMembers(objectToReturn);
        return objectToReturn;
      }

      public Object[] gatherArgsForConstructor(
          AssistedConstructor<?> constructor,
          Object[] factoryArgs) {
        int numParams = constructor.getAllParameters().size();
        int argPosition = 0;
        Object[] result = new Object[numParams];
        
        for (int i = 0; i < numParams; i++) {
          Parameter parameter = constructor.getAllParameters().get(i);
          if (parameter.isProvidedByFactory()) {
            result[i] = factoryArgs[argPosition];
            argPosition++;
          } else {
            result[i] = parameter.getValue(injector);
          }
        }
        return result;
      }
    };

    return (F) Proxy.newProxyInstance(factoryType.getClassLoader(),
        new Class[] {factoryType}, invocationHandler);
  }
}
