package assistedinject;

import com.google.inject.*;

import java.lang.annotation.Annotation;
import java.lang.reflect.Type;
import java.lang.reflect.ParameterizedType;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

/**
 * Models a method or constructor parameter.
 *
 * @author <a href="jmourits@gmail.com">Jerome Mourits</a>
 * @author <a href="jesse@swank.ca">Jesse Wilson</a>
 */
class Parameter {

  private final Type type;
  private final boolean isAssisted;
  private final Annotation bindingAnnotation;
  private final boolean isProvider;

  public Parameter(Type type, Annotation[] annotations) {
    this.type = type;
    this.bindingAnnotation = getBindingAnnotation(annotations);
    this.isAssisted = hasAssistedAnnotation(annotations);
    this.isProvider = isProvider(type);
  }

  public boolean isProvidedByFactory() {
    return isAssisted;
  }

  public TypeLiteral getTypeLiteral() {
    return TypeLiteral.get(type);
  }

  @Override
  public String toString() {
    StringBuilder result = new StringBuilder();
    if (isAssisted) {
      result.append("@Assisted");
      result.append(" ");
    }
    if (bindingAnnotation != null) {
      result.append(bindingAnnotation.toString());
      result.append(" ");
    }
    result.append(type.toString());
    return result.toString();
  }

  private boolean hasAssistedAnnotation(Annotation[] annotations) {
    for (Annotation annotation : annotations) {
      if (annotation.annotationType().equals(Assisted.class)) {
        return true;
      }
    }
    return false;
  }

  /**
   * Returns the Guice {@link Key} for this parameter.
   */
  public Object getValue(Injector injector) {
    return isProvider
        ? injector.getProvider(getBindingForType(getProvidedType(type)))
        : injector.getInstance(getPrimaryBindingKey());
  }

  private static final Map<Type, Type> PRIMITIVE_COUNTERPARTS;
  static {
    Map<Type, Type> primitiveToWrapper = new HashMap<Type, Type>() {{
        put(int.class, Integer.class);
        put(long.class, Long.class);
        put(boolean.class, Boolean.class);
        put(byte.class, Byte.class);
        put(short.class, Short.class);
        put(float.class, Float.class);
        put(double.class, Double.class);
        put(char.class, Character.class);
    }};

    Map<Type, Type> counterparts = new HashMap<Type, Type>();
    for (Map.Entry<Type, Type> entry : primitiveToWrapper.entrySet()) {
      Type key = entry.getKey();
      Type value = entry.getValue();
      counterparts.put(key, value);
      counterparts.put(value, key);
    }

    PRIMITIVE_COUNTERPARTS = Collections.unmodifiableMap(counterparts);
  }

  public boolean isBound(Injector injector) {
    return injector.getBinding(getPrimaryBindingKey()) != null
        || injector.getBinding(getAlternateGuiceBindingKey()) != null;
  }

  private Key<?> getPrimaryBindingKey() {
    return isProvider
        ? getBindingForType(getProvidedType(type))
        : getBindingForType(type);
  }

  private Key<?> getAlternateGuiceBindingKey() {
    Type counterpart = (PRIMITIVE_COUNTERPARTS.containsKey(type))
      ? PRIMITIVE_COUNTERPARTS.get(type)
      : type;
    return isProvider
        ? getBindingForType(getProvidedType(counterpart))
        : getBindingForType(counterpart);
  }


  private Type getProvidedType(Type type) {
    return ((ParameterizedType)type).getActualTypeArguments()[0];
  }

  private boolean isProvider(Type type) {
    return type instanceof ParameterizedType
        && ((ParameterizedType)type).getRawType() == Provider.class;
  }

  private Key<?> getBindingForType(Type type) {
    return bindingAnnotation != null
        ? Key.get(type, bindingAnnotation)
        : Key.get(type);
  }

  /**
   * Returns the first binding annotation from the specified list, or
   * {@code null} if there are none.
   *
   * @throws IllegalStateException if multiple binding annotations exist.
   */
  private Annotation getBindingAnnotation(Annotation[] annotations) {
    Annotation bindingAnnotation = null;
    for (Annotation a : annotations) {
      if (a.annotationType().getAnnotation(BindingAnnotation.class) != null) {
        if (bindingAnnotation != null) {
          throw new IllegalArgumentException(String.format("Parameter has " +
              "multiple binding annotations: %s and %s", bindingAnnotation, a));
        }
        bindingAnnotation = a;
      }
    }
    return bindingAnnotation;
  }
}